/*
 * Copyright (c) 2016-2020, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *    ======== spibme680.c ========
 */
#include <stdint.h>
#include <stddef.h>
#include <string.h>

/* POSIX Header files */
#include <unistd.h>

/* Driver Header files */
#include <ti/drivers/SPI.h>
#include <ti/display/Display.h>

//New
#include <ti/drivers/GPIO.h>

/* Module Header */
#include "bme680.h"

/* Driver configuration */
#include "ti_drivers_config.h"


static Display_Handle display;

//CHANGES
#define SPI_MSG_LENGTH  (256)
unsigned char masterRxBuffer[SPI_MSG_LENGTH];
unsigned char masterTxBuffer[SPI_MSG_LENGTH];

SPI_Handle      masterSpi;
SPI_Params      spiParams;
SPI_Transaction transaction;

/*BME680 Templates for function pointers*/

/* Return control or wait, for a period amount of milliseconds */
void user_delay_ms(uint32_t period)
{
    usleep(period * 1000);
}

/*
 * The parameter dev_id can be used as a variable to select which Chip Select pin has
 * to be set low to activate the relevant device on the SPI bus

     * Data on the bus should be like
     * |----------------+---------------------+-------------|
     * | MOSI           | MISO                | Chip Select |
     * |----------------+---------------------|-------------|
     * | (don't care)   | (don't care)        | HIGH        |
     * | (reg_addr)     | (don't care)        | LOW         |
     * | (don't care)   | (reg_data[0])       | LOW         |
     * | (....)         | (....)              | LOW         |
     * | (don't care)   | (reg_data[len - 1]) | LOW         |
     * | (don't care)   | (don't care)        | HIGH        |
     * |----------------+---------------------|-------------|
*/
int8_t user_spi_read(uint8_t dev_id, uint8_t reg_addr, uint8_t *reg_data, uint16_t len)
{
    int8_t rslt = 0; /* Return 0 for Success, non-zero for failure */
    SPI_Transaction transaction;
    bool transferOK;

    transaction.count = len + 1;
    transaction.rxBuf = masterRxBuffer;
    transaction.txBuf = masterTxBuffer;

    masterTxBuffer[0] = reg_addr;

    transferOK = SPI_transfer(masterSpi, &transaction);
    if (transferOK) {
        rslt = 0;
    }
    else {
        rslt = -1;
    }

    memcpy(reg_data, masterRxBuffer+1, len);

    return rslt;
}


/*
 * The parameter dev_id can be used as a variable to select which Chip Select pin has
 * to be set low to activate the relevant device on the SPI bus

 * Data on the bus should be like
 * |---------------------+--------------+-------------|
 * | MOSI                | MISO         | Chip Select |
 * |---------------------+--------------|-------------|
 * | (don't care)        | (don't care) | HIGH        |
 * | (reg_addr)          | (don't care) | LOW         |
 * | (reg_data[0])       | (don't care) | LOW         |
 * | (....)              | (....)       | LOW         |
 * | (reg_data[len - 1]) | (don't care) | LOW         |
 * | (don't care)        | (don't care) | HIGH        |
 * |---------------------+--------------|-------------|
*/
int8_t user_spi_write(uint8_t dev_id, uint8_t reg_addr, uint8_t *reg_data, uint16_t len)
{
    int8_t rslt = 0; /* Return 0 for Success, non-zero for failure */
    SPI_Transaction transaction;
    bool transferOK;
//    unsigned char masterTxBuffer[SPI_MSG_LENGTH];

    transaction.count = len + 1;
    transaction.rxBuf = masterRxBuffer;
    transaction.txBuf = masterTxBuffer;

    masterTxBuffer[0] = reg_addr;
    memcpy(masterTxBuffer+1, reg_data, len);

    transferOK = SPI_transfer(masterSpi, &transaction);
    if (transferOK) {
        rslt = 0;
    }
    else {
        rslt = -1;
    }

    return rslt;
}


/*BME680------*/

/*BME680 Initializing the sensor */

struct bme680_dev gas_sensor;
    int8_t rslt = BME680_OK;


int32_t getIAQ(struct bme680_field_data* data)
{
/*
    Calculate IAQ indices scaled 0-100% (100% is excellent) and 0-500 where 0 is excellent and 500 Hazardous
    The index is a hybrid of humidity and adverse gas content. See the Redme for further detaisl of the index.
    This software, the ideas and concepts is Copyright (c) David Bird 2019. All rights to this software are reserved.
    The result of combining humidity and gas measurements into an in index scaled from 0-100% (where 100% is good).
    This is then scaled again from 0-500
*/
    float hum_weighting = 0.25; // so hum effect is 25% of the total air quality score
    float gas_weighting = 0.75; // so gas effect is 75% of the total air quality score

     //Calculate humidity contribution to IAQ index
      float current_humidity = (float)(data->humidity) / 1000;
      float hum_reference = 40;
      float hum_score;

      if (current_humidity >= 38 && current_humidity <= 42) // Humidity +/-5% around optimum
          hum_score = 0.25 * 100;
      else
      { // Humidity is sub-optimal
        if (current_humidity < 38)
            hum_score = 0.25 / hum_reference * current_humidity * 100;
        else
        {
            hum_score = ((-0.25 / (100 - hum_reference) * current_humidity) + 0.416666) * 100;
        }
      }

      //Calculate temperature contribution to IAQ
      float current_gas = data->gas_resistance;
      float gas_lower_limit = 10000;   // Bad air quality limit  5000
      float gas_upper_limit = 1050000;  // Good air quality limit 55000  650000

      if (current_gas > gas_upper_limit) current_gas = gas_upper_limit;
      if (current_gas < gas_lower_limit) current_gas = gas_lower_limit;

      //Calculate gas contribution to IAQ index
      float gas_score = (0.75 / (gas_upper_limit - gas_lower_limit) * current_gas - (gas_lower_limit * (0.75 / (gas_upper_limit - gas_lower_limit)))) * 100.00;

      // Calculate IAQ index
      float air_quality_score = hum_score + gas_score;  // + tmp_score
      int32_t score =  ( int32_t)((100 - air_quality_score) * 5.00);

      Display_print3(display, 0, 0, "Hum_IAQ: %4.2f,  Gas_IAQ: %4.2f,  IAQ: %d ",
                     ((hum_weighting*100)-hum_score)*5, ((gas_weighting*100)-gas_score)*5, score);

      char IAQ_text[44] = "Air quality is:  ";
      if      (score >= 301)                  strcat(IAQ_text, "Hazardous");
      else if (score >= 201 && score <= 300 ) strcat(IAQ_text, "Very Unhealthy");
      else if (score >= 176 && score <= 200 ) strcat(IAQ_text, "Unhealthy");
      else if (score >= 151 && score <= 175 ) strcat(IAQ_text, "Unhealthy for Sensitive Groups");
      else if (score >=  51 && score <= 150 ) strcat(IAQ_text, "Moderate");
      else if (score >=  00 && score <=  50 ) strcat(IAQ_text, "Good");

      Display_print0(display, 0, 0, IAQ_text);

      return score;
}

/*
 *  ======== mainThread ========
 */
void *mainThread(void *arg0)
{
    uint32_t gas_resistance_max = 0;
    uint32_t gas_resistance_min = 2000000;
    int8_t first = 1;

    /* Call driver init functions */
    SPI_init();

    /* Open the HOST display for output */
    display = Display_open(Display_Type_UART, NULL);
    if (display == NULL) {
        while (1);
    }
    Display_print0(display, 0, 0, "Starting the spibme680 sensor example...\n");

    /* Open SPI as master */
    SPI_Params_init(&spiParams);
    spiParams.frameFormat = SPI_POL1_PHA1;
    spiParams.transferMode = SPI_MODE_BLOCKING;
    spiParams.mode = SPI_MASTER;
    spiParams.bitRate = 100000;
    spiParams.dataSize = 8;
    masterSpi = SPI_open(CONFIG_SPI_0, &spiParams);
    if (masterSpi == NULL) {
        Display_print0(display, 0, 0, "Error Initializing SPI");
        while (1);
    }
    else {
        Display_print0(display, 0, 0, "SPI Initialized!");
    }


    /*
     * BME680 Initializing the sensor for SPI
     */
    gas_sensor.dev_id = 0;
    gas_sensor.intf = BME680_SPI_INTF;
    gas_sensor.read = user_spi_read;
    gas_sensor.write = user_spi_write;
    gas_sensor.delay_ms = user_delay_ms;
    /* amb_temp can be set to 25 prior to configuring the gas sensor
     * or by performing a few temperature readings without operating the gas sensor.
     */
    gas_sensor.amb_temp = 25;

    /******************************/
    /*   BME680 sensor selftest   */
/*    rslt = bme680_self_test(&gas_sensor);
    if (rslt == BME680_OK)
        Display_print0(display, 0, 0, "BME680 sensor selftest OK!");
    else
        Display_print0(display, 0, 0, "BME680 sensor selftest Error");
*/

    int8_t rslt = BME680_OK;
    /* This API is the entry point.
    * It reads the chip-id and calibration data from the sensor.    */
    rslt = bme680_init(&gas_sensor);

    Display_print2(display, 0, 0,"Chip ID: 0x%x,  Chip Addr: 0x%x",gas_sensor.chip_id, gas_sensor.dev_id);
    if (rslt == BME680_OK)
        Display_print0(display, 0, 0, "BME680 Initialization OK!");
    else
        Display_print0(display, 0, 0, "BME680 Initialization Error");

    /*
     * BME680 Configuring the sensor
     */
    uint8_t set_required_settings;

    /* Set the temperature, pressure and humidity settings */
    gas_sensor.tph_sett.os_hum = BME680_OS_2X;
    gas_sensor.tph_sett.os_pres = BME680_OS_4X;
    gas_sensor.tph_sett.os_temp = BME680_OS_8X;
    gas_sensor.tph_sett.filter = BME680_FILTER_SIZE_3;

    /* Set the remaining gas sensor settings and link the heating profile */
    gas_sensor.gas_sett.run_gas = BME680_ENABLE_GAS_MEAS;
    /* Create a ramp heat waveform in 3 steps */
    gas_sensor.gas_sett.heatr_temp = 320; /* degree Celsius */
    gas_sensor.gas_sett.heatr_dur = 150; /* milliseconds */

    /* Select the power mode */
    /* Must be set before writing the sensor configuration */
    gas_sensor.power_mode = BME680_FORCED_MODE;

    /* Set the required sensor settings needed */
    set_required_settings = BME680_OST_SEL | BME680_OSP_SEL | BME680_OSH_SEL | BME680_FILTER_SEL
        | BME680_GAS_SENSOR_SEL;

    /* Set the desired sensor configuration */
    /* This API is used to set the oversampling, filter and T,P,H, gas selection
    * settings in the sensor. */
    rslt = bme680_set_sensor_settings(set_required_settings,&gas_sensor);
    if (rslt == BME680_OK)
        Display_print0(display, 0, 0, "BME680 sensor configuration OK!");
    else
        Display_print0(display, 0, 0, "BME680 sensor configuration Error");


    /* Trigger the measurement  */
    rslt = bme680_set_sensor_mode(&gas_sensor);
    if (rslt == BME680_OK)
        Display_print0(display, 0, 0, "BME680 Trigger the measurement OK!");
    else
        Display_print0(display, 0, 0, "BME680 Trigger the measurement Error");

   /* Get the total measurement duration so as to sleep or wait till the
     * measurement is complete */
    uint16_t meas_period; /* Wake up duration in ms */
    bme680_get_profile_dur(&meas_period, &gas_sensor);

    Display_print2(display, 0, 0,"Heater temperature: %d Celcius degree,   Heater duration: %d ms",gas_sensor.gas_sett.heatr_temp, gas_sensor.gas_sett.heatr_dur);
    Display_print1(display, 0, 0,"Ambient temperature: %d",gas_sensor.amb_temp);
    Display_print1(display, 0, 0,"Measurement duration: %d ms \n",meas_period);

    user_delay_ms(meas_period); /* Delay till the measurement is ready */

    struct bme680_field_data data; /*Sensor field data structure */

    /*  BME680 Reading all sensor data  */
    while(1)
    {
        /* This API reads the pressure, temperature and humidity and gas data
        * from the sensor, compensates the data and store it in the bme680_data structure */
        rslt = bme680_get_sensor_data(&data, &gas_sensor);
        if (rslt != BME680_OK)
            Display_print0(display, 0, 0, "BME680 get data Error");

        if(data.status & BME680_GASM_VALID_MSK)
        {
            if (first == 0)
            {
                if (data.gas_resistance > gas_resistance_max)
                    gas_resistance_max = data.gas_resistance;
                if (data.gas_resistance < gas_resistance_min)
                    gas_resistance_min = data.gas_resistance;
            }
            else first = 0;

//            Display_print5(display, 0, 0, "Pressure: %4.2f hPa,  Temp: %4.2f DegC,  Humidity: %4.3f %%RH,  Gas_res: %u Ohms,  Gas_res_max: %u Ohms",
//                           ((float )data.pressure)/100, ((float )data.temperature)/100, ((float )data.humidity)/1000, data.gas_resistance,  gas_resistance_max);
            Display_print3(display, 0, 0, "Pressure: %4.2f hPa,  Temp: %4.2f DegC,  Humidity: %4.3f %%RH",
                           ((float )data.pressure)/100, ((float )data.temperature)/100, ((float )data.humidity)/1000);
            Display_print4(display, 0, 0, "Gas_res: %u Ohms,  Gas_res_min: %u Ohms,  Gas_res_max: %u Ohms,  Gas_res_%%: %4.3f %%",
                           data.gas_resistance,  gas_resistance_min,  gas_resistance_max, ((float )(data.gas_resistance - gas_resistance_min) / (float )(gas_resistance_max - gas_resistance_min)) * 100.0 );
           getIAQ(&data);
        }
        else
            /* Avoid using measurements from an unstable heating setup */
            Display_print0(display, 0, 0, "FASM Valid mask Error");

        /* Trigger the next measurement if you would like to read data out continuously */
        if (gas_sensor.power_mode == BME680_FORCED_MODE) {
               rslt = bme680_set_sensor_mode(&gas_sensor);
               if (rslt != BME680_OK)
                   Display_print0(display, 0, 0, "Trigger the next measurement Error");
        }
        sleep(1);
    }
}
